import datetime
from abc import ABCMeta, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import requests
from cachier import cachier  # type: ignore
from loguru import logger
from packaging.version import LegacyVersion, Version, parse
from requests.models import Response
from rex import rex  # type: ignore

TVer = Union[LegacyVersion, Version]
TVerList = List[TVer]
TFileNameList = List[str]
TDictComponent = Dict[str, Union[Optional[str], TVer, TFileNameList, List[TVer]]]


class ComponentType(Enum):
    DOCKER = "docker-image"
    PYPI = "pypi"


class Component(metaclass=ABCMeta):

    DEFAULT_PREFIX: Optional[str] = None
    DEFAULT_FILTER: str = "/.*/"
    DEFAULT_FILES: TFileNameList = []
    DEFAULT_EXLUDE_VERSIONS: List[str] = []
    DEFAULT_REPO: Optional[str] = None
    LATEST_TAGS: List[str] = ["latest"]
    DEFAULT_VERSION_PATTERN: str = "{version}"

    def __init__(
        self,
        component_type: ComponentType,
        component_name: str,
        current_version_tag: str,
    ) -> None:
        self.component_type = component_type
        self.component_name: str = component_name
        self.current_version_tag: str = current_version_tag
        self.current_version: TVer = parse(current_version_tag)
        self.version_tags: List[str] = []
        self.next_version: TVer = self.current_version
        self.next_version_tag: str = self.current_version_tag
        self.prefix: Optional[str] = self.DEFAULT_PREFIX
        self.filter: str = self.DEFAULT_FILTER
        self.files: TFileNameList = self.DEFAULT_FILES
        self.exclude_versions: List[str] = self.DEFAULT_EXLUDE_VERSIONS
        self.version_pattern = self.DEFAULT_VERSION_PATTERN
        super().__init__()

    def newer_version_exists(self) -> bool:
        if self.current_version_tag in self.LATEST_TAGS:
            return False
        else:
            return self.next_version > self.current_version

    @abstractmethod
    def fetch_versions_tags(self) -> List[str]:
        """should return a list of versions eg.: ('1.0.1', '2.0.2')"""
        pass  # pragma: no cover

    # TODO move max statement after self.next_version= to new mehtod: get_max_version_number()
    def check(self) -> bool:
        if self.current_version_tag not in self.LATEST_TAGS:
            self.version_tags = self.fetch_versions_tags()

            self.next_version = max(
                [
                    parse(tag)
                    for tag in self.version_tags
                    if (tag == rex(self.filter)) and tag not in self.exclude_versions
                ]
            )
            self.next_version_tag = f"{(self.prefix or '')}{str(self.next_version)}"

        return self.newer_version_exists()

    def to_dict(self) -> TDictComponent:
        ret: TDictComponent = {
            "component-type": self.component_type.value,
            "current-version": self.current_version_tag,
            "next-version": self.next_version_tag,
        }

        if self.prefix != self.DEFAULT_PREFIX:
            ret["prefix"] = self.prefix
        if self.filter != self.DEFAULT_FILTER:
            ret["filter"] = self.filter
        if self.files != self.DEFAULT_FILES:
            ret["files"] = self.files
        if self.exclude_versions != self.DEFAULT_EXLUDE_VERSIONS:
            ret["exclude-versions"] = self.exclude_versions
        if self.version_pattern != self.DEFAULT_VERSION_PATTERN:
            ret["version-pattern"] = self.version_pattern
        return ret

    def name_version_tag(self, version_tag: str) -> str:
        d: Dict[str, str] = {"version": version_tag, "component": self.component_name}
        return self.version_pattern.format(**d)

    def count_occurence(self, string_to_search: str) -> int:
        return string_to_search.count(self.name_version_tag(self.current_version_tag))

    def replace(self, string_to_replace: str) -> str:
        return string_to_replace.replace(
            self.name_version_tag(self.current_version_tag),
            self.name_version_tag(self.next_version_tag),
        )

    def update_files(self, base_dir: Optional[Path], dry_run: bool = False) -> int:
        counter: int = 0
        if base_dir is None:
            raise FileNotFoundError("base_dir is None")
        for file_name in self.files:
            file = Path(base_dir / file_name)
            orig_content: str = file.read_text()
            if self.count_occurence(orig_content) > 1:
                logger.error(
                    f"Too many versions of {self.current_version_tag} occurence in {orig_content}!"
                )
                raise Exception(
                    f"Too many versions of {self.current_version_tag} occurence in {orig_content}!"
                )

            if not dry_run:
                new_content: str = self.replace(orig_content)
                if new_content == orig_content:
                    logger.error(
                        (
                            f"Error in version replacment for {self.component_name}: "
                            f"no replacement done for current_version"
                            f": {self.name_version_tag(self.current_version_tag)} "
                            f"and next_version: {self.name_version_tag(self.next_version_tag)} "
                            f"in file: {str(file)}"
                        )
                    )
                    raise Exception(
                        f"Error in version replacment for {self.component_name}: no replacement done for current_version"
                    )
                file.write_text(new_content)
            counter += 1
        return counter


def clear_versions_cache() -> None:
    fetch_docker_images_versions.clear_cache()  # type: ignore
    fetch_pypi_versions.clear_cache()  # type: ignore


@cachier(stale_after=datetime.timedelta(days=3))  # type: ignore[misc]
def fetch_docker_images_versions(
    repo_name: str, component_name: str, token_url: Optional[str] = None
) -> List[str]:
    logger.info(f"{repo_name}:{component_name} - NOT CACHED")
    payload: Dict[str, str] = {
        "service": "registry.docker.io",
        "scope": f"repository:{repo_name}/{component_name}:pull",
    }
    r: Response = requests.get(
        token_url or DockerImageComponent.TOKEN_URL, params=payload
    )
    if not r.status_code == 200:
        print(f"Error status {r.status_code}")
        raise Exception("Could not get auth token")

    token: str = r.json()["token"]
    h = {"Authorization": f"Bearer {token}"}
    r = requests.get(
        f"https://index.docker.io/v2/{repo_name}/{component_name}/tags/list", headers=h
    )
    ret: List[str] = r.json().get("tags", [])
    return ret


@cachier(stale_after=datetime.timedelta(days=3))  # type: ignore[misc]
def fetch_pypi_versions(component_name: str) -> List[str]:
    r: Response = requests.get(f"https://pypi.org/pypi/{component_name}/json")
    # it returns 404 if there is no such a package
    if not r.status_code == 200:
        return list()
    else:
        return list(r.json().get("releases", {}).keys())


class DockerImageComponent(Component):

    DEFAULT_VERSION_PATTERN: str = "{component}:{version}"
    TOKEN_URL: str = "https://auth.docker.io/token"

    def __init__(
        self, repo_name: str, component_name: str, current_version_tag: str
    ) -> None:
        super(DockerImageComponent, self).__init__(
            ComponentType.DOCKER, component_name, current_version_tag
        )
        self.repo_name = repo_name
        self.version_pattern = self.DEFAULT_VERSION_PATTERN

    def fetch_versions_tags(self) -> List[str]:
        return fetch_docker_images_versions(self.repo_name, self.component_name)

    def to_dict(self) -> TDictComponent:
        ret: TDictComponent = super(DockerImageComponent, self).to_dict()
        ret["docker-repo"] = self.repo_name
        return ret


class PypiComponent(Component):

    DEFAULT_VERSION_PATTERN: str = "{component}=={version}"

    def __init__(
        self, component_name: str, current_version_tag: str, **_ignored: Any
    ) -> None:
        super(PypiComponent, self).__init__(
            ComponentType.PYPI, component_name, current_version_tag
        )
        self.version_pattern = self.DEFAULT_VERSION_PATTERN

    def fetch_versions_tags(self) -> List[str]:
        return fetch_pypi_versions(self.component_name)


class ComponentFactory:
    def get(self, component_type: str, **args: Any) -> Component:
        if component_type == ComponentType.DOCKER.value:
            return DockerImageComponent(**args)
        elif component_type == ComponentType.PYPI.value:
            return PypiComponent(**args)
        else:
            raise ValueError(f"Componet type: {component_type} :not implemented!")


factory = ComponentFactory()



import copy
import datetime
from pathlib import Path
import pprint
from subprocess import run
from typing import Any, Dict, List, Optional, Tuple
import pkg_resources

import click
import yaml
from loguru import logger
from plumbum import local  # type: ignore

from updater import TPlumbumRunReturn, git_check, plumbum_msg, components
from updater.components import ComponentType, Component


class Config:

    STATE_FILES_UPDATED = "FILES_UPDATED"
    STATE_TEST_RUN = "TEST_RUN"
    STATE_CONFIG_SAVED = "CONFIG_SAVED"
    STATE_COMMITED_CHANGES = "COMMITED_CHANGES"
    STATE_UPDATE_STARTED = "UPDATE_STARTED"
    STATE_UPDATE_SKIPPED = "UPDATE_SKIPPED"
    STATE_UPDATE_DONE = "UPDATE_DONE"

    def __init__(self, components_yaml_file: Optional[Path] = None) -> None:
        self.components: List[Component] = []
        self.config_file = components_yaml_file
        self.project_dir: Optional[
            Path
        ] = self.config_file.parent if self.config_file else None
        self.test_command: Optional[List[str]] = None
        self.test_dir: Optional[Path] = None
        self.git_commit: bool = True
        self.status: Dict[str, Dict[str, str]] = {}

    def update_status(self, component: Component, step: str) -> None:
        if component.component_name not in self.status:
            self.status[component.component_name] = {}
        comp = self.status[component.component_name]
        if step in [self.STATE_UPDATE_STARTED, self.STATE_UPDATE_SKIPPED]:
            message = f"{step} for {component.component_name} in version {component.current_version_tag}"
        elif step in [self.STATE_UPDATE_DONE]:
            message = f"{step} for {component.component_name} in version {component.next_version_tag}"
        else:
            message = step
        comp[str(datetime.datetime.now())] = message

    def get_status(self) -> str:
        return pprint.pformat(self.status, indent=4)

    def add(self, component: Component) -> int:
        self.components.append(component)
        return self.components.index(self.components[-1])

    def components_to_dict(self) -> Dict[str, components.TDictComponent]:
        return {
            component.component_name: component.to_dict()
            for component in self.components
        }

    def save_to_yaml(self, file: Optional[str] = None) -> None:
        path = file or self.config_file
        if path:
            file_to_save: Path = Path(path)
            yaml.dump(self.components_to_dict(), open(file_to_save, "w"))
        else:
            logger.error(
                f"No config file provided and no config file found in project directory."
            )
            raise FileNotFoundError(
                "No config file provided and no config file found in project directory."
            )

    def save_config(
        self,
        destination_file: Optional[str] = None,
        dry_run: bool = False,
        print_yaml: bool = False,
    ) -> None:
        if not dry_run:
            if destination_file:
                self.save_to_yaml(destination_file)
            elif self.config_file:
                self.save_to_yaml()

        if print_yaml:
            click.echo(pprint.pformat(yaml.dump(self.components_to_dict()), indent=4))

    def read_from_yaml(self, file: Optional[Path] = None) -> None:
        read_file = file or self.config_file
        self.components = []

        components_dict: Dict[str, Any] = (
            yaml.safe_load(open(read_file)) if read_file and read_file.is_file() else {}
        ) or {}

        for component_name in components_dict:
            compd = components_dict[component_name]
            params: components.TDictComponent = {
                "component_name": component_name,
                "current_version_tag": compd["current-version"],
                "repo_name": compd.get("docker-repo", Component.DEFAULT_REPO),
            }
            last_index = self.add(
                components.factory.get(str(compd["component-type"]), **params)
            )
            comp = self.components[last_index]
            comp.prefix = compd.get("prefix", comp.DEFAULT_PREFIX)
            comp.filter = compd.get("filter", comp.DEFAULT_FILTER)
            comp.files = compd.get("files", comp.DEFAULT_FILES)
            comp.exclude_versions = compd.get(
                "exclude-versions", comp.DEFAULT_EXLUDE_VERSIONS
            )
            comp.version_pattern = compd.get(
                "version-pattern", comp.DEFAULT_VERSION_PATTERN
            )

    def add_from_requirements(self, req_file: str, req_source: str) -> None:

        file_to_read = Path(req_file)
        if not file_to_read.is_file():
            logger.error(f"File {file_to_read} not found.")
            raise FileNotFoundError(f"Missing file for import: {file_to_read}")
        with file_to_read.open() as requirements_txt:
            for requirement in pkg_resources.parse_requirements(requirements_txt):
                if len(requirement.specs) == 0 or requirement.specs[0][0] != "==":
                    print(
                        f"{requirement.project_name} has incompatible version specifier: {requirement.specs}"
                    )
                    continue

                name, version = (requirement.project_name, requirement.specs[0][1])
                if not any(
                    x.component_name == name and x.component_type == ComponentType.PYPI
                    for x in self.components
                ):
                    self.add(
                        components.factory.get(
                            component_type=ComponentType.PYPI.value,
                            component_name=name,
                            current_version_tag=version,
                        )
                    )
                    comp = self.components[-1]
                    if req_source == "pipfile":
                        comp.version_pattern = '{component} = "=={version}"'
                        comp.files = ["Pipfile"]
                    elif req_source == "requirements":
                        comp.version_pattern = "{component}=={version}"
                        comp.files = ["requirements.txt"]
                    comp.filter = (
                        "/^" + (version.count(".")) * "\\d+\\." + "\\d+$/"
                    )  # due ot backslash being forbidden in f-strings inside curly braces

    def count_components_to_update(self) -> int:
        self.check()
        return sum(
            [1 for component in self.components if component.newer_version_exists()]
        )

    def check(self) -> List[Tuple[str, bool]]:
        return [(comp.component_name, comp.check()) for comp in self.components]

    def run_tests(self, processed_component: Component) -> None:
        assert self.test_command, "No test command provided."
        ret = run(self.test_command, cwd=(self.test_dir or self.project_dir))
        if ret.returncode != 0:
            raise ValueError(
                f'{click.style("Error!", fg="red")} ( {processed_component.component_name} ) {str(ret)}'
            )

    def commit_changes(
        self, component: Component, from_version: str, to_version: str, dry_run: bool
    ) -> None:
        git = local["git"]
        assert self.config_file, "No config file found."
        # TODO shouldn't it be self.project_dir param rather?
        with local.cwd(self.config_file.parent):
            ret: TPlumbumRunReturn = git_check(
                git["diff", "--name-only"].run(retcode=None)
            )
            changed_files: List[str] = ret[1].splitlines()
            assert set(component.files).issubset(
                set(changed_files)
            ), f"Not all SRC files are in git changed files.\n{plumbum_msg(ret)}"
            if not dry_run:
                git_check(git["add", self.config_file.name].run(retcode=None))
                for file_name in component.files:
                    git_check(git["add", file_name].run(retcode=None))
                commit_message = (
                    f"{component.component_name} "
                    f"updated from: {from_version} to: {to_version}"
                )
                git_check(
                    git["commit", f"--message={commit_message}"].run(retcode=None)
                )

    def update_files(self, dry_run: bool = False) -> Tuple[int, int]:
        file_counter = 0
        components_counter = 0
        for component in self.components:
            if component.newer_version_exists():
                orig_current_tag = component.current_version_tag
                orig_next_tag = component.next_version_tag
                self.update_status(component, self.STATE_UPDATE_STARTED)
                prev_file_counter = file_counter
                file_counter += component.update_files(self.project_dir, dry_run)
                components_counter += 1 if file_counter > prev_file_counter else 0
                self.update_status(component, self.STATE_FILES_UPDATED)
                if self.test_command:
                    self.run_tests(component)
                    self.update_status(component, self.STATE_TEST_RUN)

                if not dry_run:
                    component.current_version = copy.deepcopy(component.next_version)
                    component.current_version_tag = copy.deepcopy(
                        component.next_version_tag
                    )
                self.save_config(dry_run=dry_run)
                self.update_status(component, self.STATE_CONFIG_SAVED)

                if self.git_commit:
                    self.commit_changes(
                        component, orig_current_tag, orig_next_tag, dry_run
                    )
                    self.update_status(component, self.STATE_COMMITED_CHANGES)
                self.update_status(component, self.STATE_UPDATE_DONE)
            else:
                self.update_status(component, self.STATE_UPDATE_SKIPPED)

        return components_counter, file_counter

    def get_versions_info(self) -> List[str]:
        new = [
            (
                f"{c.component_name} - current: {c.current_version_tag} "
                f"next: {click.style(c.next_version_tag, fg='green')}"
            )
            for c in self.components
            if c.newer_version_exists()
        ]
        new.sort()
        return new




from typing import Tuple


TPlumbumRunReturn = Tuple[int, str, str]


def plumbum_msg(command_exit: TPlumbumRunReturn) -> str:
    return (
        f"Exit code {command_exit[0]}.\n"
        f"Command output:\n{command_exit[1]}."
        f"\nError msg:\n{command_exit[2]}\n"
    )


def git_check(ret: TPlumbumRunReturn) -> TPlumbumRunReturn:
    assert ret[0] == 0, f"Error returned by git.\n{plumbum_msg(ret)}"
    return ret



from pathlib import Path
import sys
from typing import List, Optional

import click
from click.core import Context
from loguru import logger
import pkg_resources

from updater import components, config_yaml


@click.group()
@click.version_option(
    version=Path(
        pkg_resources.resource_filename(__name__, "updater/VERSION")
    ).read_text()
)  # type: ignore
@click.option(
    "--file",
    type=click.Path(exists=True, writable=True),
    help="YAML file with components configuration. If not present other options for 'check' command are required.",
)  # type: ignore
@click.option(
    "--destination-file",
    "destination_file",
    type=click.Path(),
    help="If this option is given components configuration with new versions will be written here.",
)  # type: ignore
@click.option(
    "--dry-run",
    "dry_run",
    is_flag=True,
    help="If set no changes to any files are written.",
)  # type: ignore
@click.option(
    "--print",
    "print_yaml",
    is_flag=True,
    help="Config is printed to stdout at the end.",
)  # type: ignore
@click.pass_context
def cli(
    ctx: Context,
    file: Optional[Path],
    destination_file: Optional[Path],
    dry_run: bool,
    print_yaml: bool,
) -> None:
    config_file: Optional[Path] = None
    if file is not None:
        config_file = Path(file).absolute()
    elif Path.cwd().absolute().joinpath("components.yaml").is_file():
        config_file = Path.cwd().absolute().joinpath("components.yaml")

    if ctx.obj is None:
        ctx.obj = {}

    ctx.obj["config"] = config_yaml.Config(components_yaml_file=config_file)
    ctx.obj["config_file"] = config_file
    ctx.obj["destination_file"] = destination_file
    ctx.obj["dry_run"] = dry_run
    ctx.obj["print_yaml"] = print_yaml


@cli.command()
@click.option(
    "--type",
    "component_type",
    help="Component type: docker-image or pypi package.",
    type=click.Choice(["docker-image", "pypi"]),
)  # type: ignore
@click.option(
    "--component", help="A component name for which the version should be verified."
)  # type: ignore
@click.option(
    "--repo_name", help="A repository name if component is a docker image."
)  # type: ignore
@click.option(
    "--version_tag",
    help="Version tag eg. v2.3.0 against which new version check will be run.",
)  # type: ignore
@click.option(
    "--verbose",
    is_flag=True,
    help="Print detailed info for each component about new version avaialble.",
)  # type: ignore
@click.option(
    "--clear-cache",
    "clear_cache",
    is_flag=True,
    help="Clear all the cached responses about versions in rpositories.",
)  # type: ignore
@click.option(
    "--ignore-default-file",
    "ignore_default_file",
    is_flag=True,
    help="Ignore components.yaml file in local directory if exists.",
)  # type: ignore
@click.pass_context
def check(
    ctx: Context,
    component_type: Optional[str],
    component: Optional[str],
    repo_name: Optional[str],
    version_tag: Optional[str],
    verbose: bool,
    clear_cache: bool,
    ignore_default_file: bool,
) -> None:
    """Check if new versions of ddefined components are available.
    """
    config = ctx.obj["config"]
    destination_file = ctx.obj["destination_file"]
    dry_run = ctx.obj["dry_run"]
    print_yaml = ctx.obj["print_yaml"]

    if clear_cache:
        components.clear_versions_cache()
        sys.exit(0)

    if ignore_default_file:
        config.config_file = None

    config.read_from_yaml()

    if component is not None:
        config.add(
            components.factory.get(
                component_type=str(component_type),
                repo_name=repo_name,
                component_name=component,
                current_version_tag=version_tag,
            )
        )

    ret_mess: List[str] = []
    ret_mess.append(f"{len(config.components)} components to check")
    ret_mess.append(f"{config.count_components_to_update()} components to update")
    config.save_config(destination_file, dry_run, print_yaml)
    if verbose:
        ret_mess.extend(config.get_versions_info())
    click.echo("\n".join(ret_mess))


@cli.command()
@click.option(
    "--test-command",
    "test_command",
    help="Command that should be run after updating each component.",
)  # type: ignore
@click.option(
    "--test-dir",
    "test_dir",
    type=click.Path(),
    help="If test-command param is given, this will be the context dir to run it.",
)  # type: ignore
@click.option(
    "--git-commit",
    "git_commit",
    is_flag=True,
    help="When set after each components update, git commit is performed in active branch.",
)  # type: ignore
@click.option(
    "--project-dir",
    "project_dir",
    type=click.Path(),
    help="If given, then it will be treated as a root dir for paths in config file.",
)  # type: ignore
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Print at the end detailed info for each component about update process.",
)  # type: ignore
@click.option(
    "-vv",
    "--very-verbose",
    is_flag=True,
    help="Print at the end detailed info for each component about update process.",
)  # type: ignore
@click.pass_context
def update(
    ctx: Context,
    test_command: Optional[str],
    test_dir: Optional[Path],
    git_commit: bool,
    project_dir: Optional[Path],
    verbose: bool,
    very_verbose: bool,
) -> None:
    """Update files, run test and commit changes."""
    config = ctx.obj["config"]
    dry_run = ctx.obj["dry_run"]
    destination_file = ctx.obj["destination_file"]
    print_yaml = ctx.obj["print_yaml"]

    config.test_command = test_command.split() if test_command is not None else None
    config.test_dir = test_dir
    config.git_commit = git_commit
    config.project_dir = Path(project_dir or config.project_dir)
    config.read_from_yaml()
    config.check()
    config.save_config(destination_file, dry_run, print_yaml)

    try:
        components_updated, files_updated = config.update_files(dry_run)
        if verbose or very_verbose:
            click.echo(
                f"{components_updated} components updated, {files_updated} files updated"
            )
        if very_verbose:
            click.echo(config.get_status())
            logger.trace(config.get_status())
    except Exception as e:
        logger.error(
            (
                f'{click.style("Something went wrong!!!", "red")}\n'
                f"Config status:\n"
                f"{config.get_status()}\n"
                f"{e}"
            )
        )
        sys.exit(2)


@cli.command()
@click.option(
    "--source",
    help="Source of the requirement.txt file.",
    type=click.Choice(["requirements", "pipfile"]),
    required=True,
)  # type: ignore
@click.option(
    "--requirements-file",
    "requirements_file",
    type=click.Path(exists=True),
    help="Requirements.txt file from which packages and versions will be added to components.yaml file.",
    required=True,
)  # type: ignore
@click.pass_context
def import_req(ctx: Context, source: str, requirements_file: Path) -> None:
    """Imports python packages from requirements.txt file."""
    config = ctx.obj["config"]
    dry_run = ctx.obj["dry_run"]
    destination_file = ctx.obj["destination_file"]
    print_yaml = ctx.obj["print_yaml"]

    config.read_from_yaml()
    config.add_from_requirements(requirements_file, source)
    config.save_config(destination_file, dry_run, print_yaml)


if __name__ == "__main__":
    cli(obj={})  # pragma: no cover
